local function strash_log(p, q) -- debug
	strash_Log(p,q,"strashBot_mods")
end


local _mods= {}
rawset(_G,"strash_mods",{
    mods= _mods
})

local isModChanceChangeAbsolute= nil


local function mod_probaChange(var, modName)
	if (not var) or var<=0 then
		if modName==nil then
			print("[strashBot_mods] - for all mods, proba will be set to 0")
			for mod,t in pairs(_mods) do
				t.auto_pick_proba= 0
				t.desired_proba= 0
			end
		else
			local t= _mods[modName]
			if t then
				print("[strashBot_mods] - for mod \""..modName.."\", proba will be set to "..0)
				t.auto_pick_proba= 0
				t.desired_proba= 0
			else
				print("[strashBot_mods] - not such mod \""..modName.."\", found...")
			end
		end
	elseif var>=50 or (isModChanceChangeAbsolute and isModChanceChangeAbsolute.value) then
		if modName==nil then
			print("[strashBot_mods] - for all mods, proba will be set to "..var)
			for mod,t in pairs(_mods) do
				t.auto_pick_proba= var
				t.desired_proba= var
			end
		else
			local t= _mods[modName]
			if t then
				print("[strashBot_mods] - for mod \""..modName.."\", proba will be set to "..var)
				t.auto_pick_proba= var
				t.desired_proba= var
			else
				print("[strashBot_mods] - no such mod \""..modName.."\", found...")
			end
		end
	else
		for mod,t in pairs(_mods) do
			local n_incomp= 0

			local total_access_rProba= 1*FRACUNIT
			for mod_,t_ in pairs(_mods) do
				if not t_.auto_pick then continue end
				if mod_==mod then break end
				local access_rProba= 1*FRACUNIT-FixedDiv(t_.auto_pick_proba*FRACUNIT,100*FRACUNIT)
				for i,incomp in pairs(t_.incompatibilities) do
					if incomp==mod then
						total_access_rProba= FixedMul(total_access_rProba,access_rProba)
						n_incomp= $+1
						break
					end
				end
			end
			local new_mod_proba= ((modName==nil or mod==modName) and var or t.desired_proba)
			if n_incomp>0 then
				-- with:
				-- W: calculated proba to achieve provided proba of mod activating
				--   taking into account previous incompatible mods
				-- G: the absolute proba desired
				-- N: number of mod beforehand that are incompatible with
				--   current mod
				-- Gk: the proba of the k-th incompatible mod beforehand
				-- we have:
				-- W= (G / ((1-G)*(1-G2)*...*(1-GN))
				local new_mod_rProba= FixedDiv(
					((modName==nil or mod==modName) and
							FixedDiv(var*FRACUNIT,100*FRACUNIT)
						or
							FixedDiv(t.desired_proba*FRACUNIT,100*FRACUNIT)
					),
					total_access_rProba
				)
				new_mod_proba= FixedInt(FixedCeil(FixedMul(new_mod_rProba,100*FRACUNIT)))	
			end
			strash_log(server, "for mod \""..mod.."\", proba will be set to "..new_mod_proba)
			t.auto_pick_proba= new_mod_proba
			t.desired_proba= var
		end
	end
end


local ANNOUNCEMENT_FLAG= {
	NEVER= 0,

	RACE_LAPS= 1,
	RACE_SECTION= 2,
	ANY_RACE= 3,

	BATTLE= 4,

	ALWAYS= 255,

	allExcept= function(f)
		return ( 255 & (~f) )
	end
}

local function add_mod(modName, enableFunc,
        disableFunc, levelChangeFunc, activationConditionFunc,
		minAutoRoundCoolDown, announcementContext,
        incomp, neverAutoPick)
    incomp= incomp or {}
	announcementContext= ($==nil) and ANNOUNCEMENT_FLAG.ALWAYS or announcementContext
	modName= modName:lower()
	minAutoRoundCoolDown= ($==nil or $<0) and 0 or $

    if not _mods[modName] then
		_mods[modName]= {
			on_func= function(p)
					enableFunc(p)
					local mod= _mods[modName]
					mod.activated= true
					mod.actual_autoRoundCoolDown= (mod.autoRoundCoolDown or 0)
				end,
			off_func= function(p)
					disableFunc(p)
					local mod= _mods[modName]
					mod.activated= false
					mod.actual_autoRoundCoolDown= ($ and $>1) and ($-1) or 0
				end,
			level_change_func= levelChangeFunc,
			activation_condition_func= activationConditionFunc,
			incompatibilities= incomp,
			activated= false,
			manually_activated= false,
			auto_pick= (neverAutoPick==nil or ((not neverAutoPick) and true or false)),
			desired_proba= (modChangeChance~=nil) and modChangeChance or 5,
			auto_pick_proba= (modChangeChance~=nil) and modChangeChance or 5,
			announcement_ctx= announcementContext,
			autoRoundCoolDown= minAutoRoundCoolDown
		}
    else
        print("Mod \""..modName.."\" already added")
    end
end
strash_mods.add_mod= add_mod
strash_mods.ANNOUNCEMENT_FLAG= ANNOUNCEMENT_FLAG

local function manuallyToggleMod(p, modName, b)
	local t= _mods[modName]
	if not t then
		print("[strashBot_mods] - no such mod \""..modName.."\", found...")
		return
	end
	if b==nil then
		b= (not t.manually_activated)
	end
	
	t.manually_activated= b
	strash_log(p, modName.."'s manualy_activation set to "..(b and "true" or "false"))
	if b then
		strash_log(p,"try manually calling "..modName.."'s on func")
		if (t.activation_condition_func==nil) or (t.activation_condition_func and t.activation_condition_func(p, true))
			t.on_func(p)
		end
	else
		strash_log(p,"try manually calling "..modName.."'s off func")
		t.off_func(p)		
	end
end
strash_mods.manuallyToggleMod= manuallyToggleMod

local function change_level_call(p)
	for mod,t in pairs(_mods) do
		if firstTo~=nil and firstTo>0 and not t.manually_activated then
			strash_log(p,"FT"..firstTo.." detected, deactivating \""..mod.."\"")
			t.off_func(p)
		end
		if t.level_change_func then
			t.level_change_func(p)
		end
	end
end
strash_mods.change_level_call= change_level_call


local modChangeChance= 7

COM_AddCommand("strash_modchange",
	function(player, val, modName)
		if val==nil then
			CONS_Printf(player,"[strashBot] - call: strash_mod number [modname]")
			return
		end

		local v= tonumber(val)
		v= (v<(-1)) and (-1) or ((v>100) and 100 or v)
		strash_log(player, "going with strash_modchange("..v..
			(modName~=nil and (", "..modName) or "")
			..")"
		)

		if v==nil then
			CONS_Printf(player,"[strashBot] - first arg of \"strash_modchange\" must be an interger value")
		else
			if modName==nil then
				modChangeChance= v
			end
			mod_probaChange(v, modName)
		end
	end,
	1
)

isModChanceChangeAbsolute= CV_RegisterVar({
	name= "strash_modchancechangeabsolute",
	defaultvalue = "On",
	flags = CV_NETVAR|CV_CALL,
	possiblevalue = CV_OnOff,
	func = function(var)
		mod_probaChange(modChangeChance)
	end
})

local function changeModChance()
	if modChangeChance <= 0 or (firstTo~=nil and firstTo>0) then return end
	if server.strash_modChange then
		strash_log(server, "Evaluating mod chance change")
		server.strash_modChange= false
		local excluded= {}
		for mod,t in pairs(_mods) do
			strash_log(server, "mod \""..mod.." \"auto_pick is "..(t.auto_pick==nil and 'nil' or (t.auto_pick and "true" or "false")))
			if not t.auto_pick then continue end
			local is_incomp= false
			for i=1,#excluded do
				if excluded[i]==mod then
					strash_log(server, "\""..excluded[i].."\" is incompatible with an activated or activating mod")
					is_incomp= true

					break
				end
			end

			if is_incomp and not t.manually_activated then
				strash_log(server, "... turning off \""..mod.."\"")
				t.manually_activated= false
				t.off_func(server)

				continue
			end

			local chance= P_RandomRange(1, 100)
			local chancePicked= (chance <= t.auto_pick_proba)
			--strash_log(server, "for \""..mod.."\", draw "..chance.."% over "..t.auto_pick_proba.."% => "..
			--	(chancePicked and "LESGO!" or "nope...")
			--)
			CONS_Printf(server, "Mod '"..mod.."' pick: "..chance.."% over "..t.auto_pick_proba.."% => " ..
				(chancePicked and "Yes" or "No")
			)
			if chancePicked or t.manually_activated then
				local incomp= t.incompatibilities or {}
				local cancel= false
				if t.activation_condition_func and (not t.activation_condition_func(server, t.manually_activated)) then
					strash_log(server, "- actually, deactivating \""..mod.."\"")
					t.manually_activated= false
					cancel= true
					t.off_func(server)
				elseif incomp and #incomp>0 then
					for i=1,#incomp do
						if (_mods[incomp[i]].manually_activated) and (not t.manually_activated) then
							strash_log(server, "- actually, deactivating \""..mod..
								"\" since it's incompatible with \""..incomp[i].."\""
							)
							t.manually_activated= false
							cancel= true
							t.off_func(server)
							break
						else
							strash_log(server, "- now excluding: "..incomp[i])
							-- concat excluded+incomp
							excluded[#excluded+1]= incomp[i]
						end
					end
				end
				if(chancePicked and not cancel) then
					if t.actual_autoRoundCoolDown and not t.manually_activated then
						strash_log(server, "can't activate '"..mod..
							"' because it's still on autoRoundCouldDown ("..
							t.actual_autoRoundCoolDown.."/"..t.autoRoundCoolDown..")"
						)
						t.off_func(server)
					else
						t.on_func(server)
					end
				end
			elseif not t.manually_activated then
				strash_log(server, "deactivating \""..mod.."\" because it wasn't manually activated")
				t.off_func(server)
			else
				t.actual_autoRoundCoolDown= t.autoRoundCoolDown or 0
			end
		end
	end
end


local mod_sounds= {}
local function setModSound(modName, sfx, sfxPlayConditionFunc, announcementContext)
	local mod= modName:lower()
	if mod_sounds and sfx then
		mod_sounds[mod]= {
			sfx= sfx,
			playCondFunc= sfxPlayConditionFunc
		}
		print("Mod '"..mod.."' was given new announcement sound")
	end
end
strash_mods.setModSound= setModSound


local function shouldModBeAnnounced(mod)
	local modflag= (mod.announcement_ctx~=nil)
						and	mod.announcement_ctx
						or	strash_mods.ANNOUNCEMENT_FLAG.ALWAYS

	if gametype==0 then
		if (not (modflag & strash_mods.ANNOUNCEMENT_FLAG.ANY_RACE))
		then
			return false
		elseif (not (modflag & strash_mods.ANNOUNCEMENT_FLAG.RACE_SECTION)) 
			and (mapheaderinfo[gamemap].levelflags & LF_SECTIONRACE)
		then
			return false
		elseif (not (mapheaderinfo[gamemap].levelflags & LF_SECTIONRACE))
			and (not (modflag & strash_mods.ANNOUNCEMENT_FLAG.RACE_LAPS))
		then
			return false
		end
	elseif gametype==1 then
		return ((modflag & strash_mods.ANNOUNCEMENT_FLAG.BATTLE) and true or false)
	end
	
	return true
end

local lastSfxPlayer_idx= -1
addHook("ThinkFrame", function()
	if mod_sounds and lastSfxPlayer_idx>=(-1)
		and leveltime>=5 and leveltime<5*TICRATE
	then
		local i= 0
		for mod,sound in pairs(mod_sounds) do
			local m= _mods[mod]
			if 	i>lastSfxPlayer_idx and
				m and m.activated and
				sound.sfx and ((not sound.playCondFunc) or sound.playCondFunc())
				and shouldModBeAnnounced(m)
			then
				strash_log(server, "startSound for mod "..mod)
				S_StartSound(nil, sound.sfx)
				lastSfxPlayer_idx= i
				break
			elseif i==lastSfxPlayer_idx and S_IdPlaying(sound.sfx) then
				break
			end
			
			i= $+1
		end
		
		-- this should only happen when no mod activated at start
		if lastSfxPlayer_idx<0 and mod_sounds["DEFAULT_SOUND"] then
			local sound= mod_sounds["DEFAULT_SOUND"]
			if sound and sound.sfx and ((not sound.playCondFunc) or sound.playCondFunc()) then
				S_StartSound(nil, sound.sfx)
				lastSfxPlayer_idx= -2
			end
		end
	end
end)

local function onServerEmpty_prepareReset()
	strash_log(server, "onServerEmpty_prepareReset")
	for mod,t in pairs(_mods) do
		t.manually_activated= false
		t.off_func(server)
		if t.level_change_func then
			t.level_change_func(server)
		end
	end
end
strash_mods.onServerEmpty_prepareReset= onServerEmpty_prepareReset


local IN_VOTE_THINKER_WAIT=16
local voteThinkerTime= 0
local voteThinkerCheck= nil
addHook("VoteThinker", function()
	voteThinkerTime= $+1
	if voteThinkerTime<IN_VOTE_THINKER_WAIT then return end

	-- all of this because, we might need to reevaluate
	-- the activation of the mods if an event like
	-- a firstTo round launch or a manual mode activation occurs
	-- during vote time (so we can't just do the calls once, at
	-- the begining)
	if(voteThinkerCheck==nil) then
		voteThinkerCheck= {}
		changeModChance()
		change_level_call(server)
		voteThinkerCheck.ft= firstTo
		for mod,t in pairs(_mods) do
			voteThinkerCheck[mod.."_enabled"]= t.activated
		end
	else
		local test= true
		if (firstTo~=voteThinkerCheck.ft) then
			test= false
			voteThinkerCheck.ft= firstTo
		else
			for mod,t in pairs(_mods) do
				test= test and (voteThinkerCheck[mod.."_enabled"]==t.activated)
				voteThinkerCheck[mod.."_enabled"]= t.activated
			end
		end
		if not test then
			change_level_call(server)
		end
	end
end)

addHook("MapLoad", function(mapnum)
	voteThinkerTime= 0
	voteThinkerCheck= nil
	lastSfxPlayer_idx= -1
end)


local function drawStringCenter(v, x, y, str, flags, tt)
	local w = v.stringWidth(str, flags, tt)
    v.drawString(x - (w / 2), y, str, flags, tt)
    return w
end


local liteColorTab= {
	"\x82", "\x83", "\x85", "\x86", "\x87"
}

hud.add(function(v,p,c)
	if(leveltime<3*TICRATE) then
		local dupadjust = {x=(v.width()/v.dupx()), y=(v.height()/v.dupy())}
		local duptweak = {x=((dupadjust.x - 320)/2), y=((dupadjust.y - 200)/2)}
		local flags= V_SNAPTOTOP | V_SNAPTOLEFT | V_ALLOWLOWERCASE
		local str=""
		local i=1
		for mod,t in pairs(_mods) do
			if t.activated and shouldModBeAnnounced(t) then
				str= str.."\x80"..(#str>0 and " - " or "")..liteColorTab[i]..
					(t.manually_activated and "*" or "")..mod
				i= (i%5)+1
			end
		end
		if firstTo~=nil and firstTo>0 then
			str= str.."\x80"..(#str>0 and " - " or "")..liteColorTab[i].."FT"..firstTo
		end
		
		if #str > 0 then
			local tx= 0
			local ty= 122+duptweak.y
			local bx= dupadjust.x
			local h= 16
			local ybrd=2

		
			v.drawFill(tx,ty,bx,h,208|flags)
			v.drawFill(tx,ty+ybrd,bx,h-2*ybrd,240|flags)
			drawStringCenter(v,160+duptweak.x,ty+h/2-5,str,flags,"thin")
		end
	end
end)


-- "NetVars" hook: https://wiki.srb2.org/wiki/Lua/Hooks#MapLoad
addHook("NetVars", function(net)
    -- all this stuff needs to be the same for everyone, yeah
	for mod,t in pairs(_mods) do
		print("NET SYNC for mod '"..mod.."'")
	 	t.auto_pick= net($)
	 	t.activated= net($)
	 	t.manually_activated= net($)
		t.auto_pick_proba= net($)
		t.desired_proba= net($)
		t.autoRoundCoolDown= net($)
		t.actual_autoRoundCoolDown= net($)
		t.cool_down= net($)
		t.base_cool_down= net($)
		t.announcement_ctx= net($)
		t.incompatibilities= net($)
	end

	lastSfxPlayer_idx= net($)
	voteThinkerTime= net($)
	voteThinkerCheck= net($)
	modChangeChance= net($)
	isModChanceChangeAbsolute= net($)
end)
